;***********************************************************************************
;12V-Akkuwchter
;---------------
;Ein 12V Bleiakku (Autobatterie) soll mit dieser Schaltung berwacht werden.
;Der Ladezustand kann grob ber 8 LEDs abgelesen werden die wie ein Bargraph 
;angeordnet sind.
;Unter 11,5 Volt blinkt die rote LED als Warnung (Tiefentladung) 
;Unter 12V leuchtet nur eine LED (rot). 
;Zwischen 12,0 und 13,8 Volt fllt sich der Bargraph.
;Ab 14,2V kann angenommen werden, das der Akku gerade geladen wird und ein whlbares
;Lauflich wird auf den LEDs abgespielt.
;Die 8 LEDs (Low Current) werden an 4 Ports nach dem 
;Charlieplexing Prinzip verschaltet und in der ISR gemuxt.
;Ein ca. 1:4 Spannungsteiler (10k, 3k3) wird zur Spanungsmessung verwendet. Vref = 5V
;-------------------------------------------------------------------------------------
;Fr ATTiny 13, 1,2 MHz interner R/C Oszilator
;Autor: Jrgen Woetzel, juergen@avr-projekte.de
;Version 1.0 vom 30.10.2013 
;***********************************************************************************
.include "tn13def.inc"
;***********************************************************************************
;Registerdefs
;***********************************************************************************
.def    ileds   =   r2
.def    tick    =   r3
.def    null    =   r4
.def	s		=	r5
.def	itemp1	=	r6
.def    itemp2  =   r7
.def    arrayl  =   r8
.def    arrayh  =   r9
.def	temp1 	= 	r16
.def    temp2   =   r17
.def    temp3   =   r18
.def    temp4   =   r19
.def    leds    =   r20
.def    ring    =   r21
;***********************************************************************************
;Programmdefs
;***********************************************************************************
.set    messen  =   0           ;Auf 1 setzen, um den Binrwert der AD-Messung anzuzeigen
.set    player  =   4           ;Einen Player whlen (0-5)
.equ    delay   =   100         ;Je hher der Wert, desto langsamer das Lauflicht (1-255)
.equ    load    =   0b11100010  ;Der AD-Wert bei 14,2 Volt
.equ    lowbat  =   0b01011000  ;Der AD-Wert bei 11,5 Volt
;-----------------------------------------------------------------------------------
;Der AD-Wert (Schwellwert) bei dem der Bargraph bis zur LED X aufleuchten soll
;-----------------------------------------------------------------------------------
;LED1 (rot) leuchtet von 11,5V - 12.0V (Wert in lowbat). Unter 11,5V blinkt LED1
.equ    led2    =   0b01110010  ;gelb   12,0 V
.equ    led3    =   0b10000001  ;gelb   12,3 V
.equ    led4    =   0b10010001  ;gelb   12,6 V
.equ    led5    =   0b10100000  ;gelb   12,9 V
.equ    led6    =   0b10110000  ;grn   13,2 V
.equ    led7    =   0b10111100  ;grn   13,5 V
.equ    led8    =   0b11001100  ;grn   13,8 V
;***********************************************************************************
;SRam
;***********************************************************************************
.dseg
array:	.byte	16
.cseg
;***********************************************************************************
;Sprungvektoren
;***********************************************************************************
.org    0
        rjmp	init
.org    OVF0addr
;***********************************************************************************
;Timerinterrupt ca. alle 1,7 mS. Ca. 73 Hz Muxfrequenz
;***********************************************************************************
        in      s,sreg              ;SREG wegspeichern
        dec     tick                ;Alle 1,7 mS (Variable fr Unterprogramm wait)
        out     ddrb,null           ;Alle Leds aus (Tristate)
        out     portb,null          ;Pullups auch aus
        lsl     ileds               ;Nchste Led ins Carry
        brne    LEDtst              ;War LED8 schon durch?
        movw    y,arrayl:arrayh     ;Wenn ja, Tabellenzeiger auf LED1
        mov     ileds,leds          ;ileds Update
        sec                         ;Zeroflag wird nach dem 8. "lsl ileds" gesetzt
        rol     ileds               ;LED1 ins Carry
LEDtst: ld      itemp1,y+           ;Daten fr Tristate und Ports der gerade aktiven LED 
        ld      itemp2,y+           ;Immer laden damit Tabellenzeiger erhht wird (Y+)
        brcc    LEDoff              ;Bei C=0, LED ausgeschaltet lassen
        out     ddrb,itemp1         ;in die C Register schreiben
        out     portb,itemp2
LEDoff: out     sreg,s              ;SREG zurck
        reti   
;***********************************************************************************
;LED Tabelle
;***********************************************************************************
        ;---------------------------------------------------------------------------
        ;Zum LED Einschalten Tabwert in DDRB - PORTB schreiben
        ;Zum Ausschalten der LED in beide Register eine null (Tristate, Pullup aus)
        ;Die Tab wird um den Z-Pointer freizuhalten ins SRam kopiert
        ;---------------------------------------------------------------------------
LEDTab: .db     0b0101,0b0100       ;LED1 einschalten, erster Wert in DDRB, 2. Wert PortB
        .db     0b0101,0b0001       ;LED2 ein usw.
        .db     0b1001,0b1000
        .db     0b1001,0b0001
        .db     0b0110,0b0100
        .db     0b0110,0b0010
        .db     0b1010,0b1000
        .db     0b1010,0b0010
;***********************************************************************************
;Initialisierung bei Power on
;***********************************************************************************
init:   ldi     temp1,LOW(RAMEND)   ;Stack
        out     SPL,temp1
        ;---------------------------------------------------------------------------
        ;Timeroverflow Interrupt anwerfen
        ;---------------------------------------------------------------------------
        ldi     temp1,1<<cs01       ;Prescaler /8
        out     TCCR0B,temp1        ;1,2MHz/256/8= 586Hz. ISR alle 1,7 mS
        ldi     temp1, 1<<toie0     ;Interrupt bei Timer Overflow ein
        out     TIMSK0, temp1
        ;---------------------------------------------------------------------------
        ;AD-Wandler
        ;---------------------------------------------------------------------------
        ldi     temp1,1<<ADC2D       ;PB4 abschalten
        out     DIDR0,temp1
        ldi     temp1,1<<ADEN|1<<ADPS0|1<<ADPS1 ;ADC ein, Prescaler=8 (1,2MHz/8=150KHz)
        out     ADCSRA,temp1
        ldi     temp1,1<<mux1          ;vref=5V, ADC2 (PB4)
        out     admux,temp1
        ;---------------------------------------------------------------------------
        ;Startwerte Programm
        ;---------------------------------------------------------------------------
        ldi     yl,low(array)       ;LED Tabelle ins SRam kopieren
        ldi     yh,high(array)
        movw    arrayl:arrayh,y     ;Zeiger fr die ISR merken
        ldi     zl,low(LEDTab*2)    ;Z auf Tabelle mit DDRB und Portpatterns
        ldi     zh,high(LEDTab*2)
        ldi     temp2,16            ;16 Werte zu kopieren
copy:   lpm     temp1,z+            ;Wert aus Tab holen
        st      y+,temp1            ;und ins SRam kopieren
        dec     temp2               ;das ganze 16 mal
        brne    copy
        clr     null                ;Konstante Null
        ldi     leds,0b00000000     ;Alle LEDs aus
        sei                         ;ISR Los
;***********************************************************************************
;AD-Wert direkt auf den 8 LEDs ausgeben
;***********************************************************************************
 .if messen ==1
 m1:    rcall   ad                  ;Messen
        mov     leds,temp1          ;AD-Wert -> leds
        rjmp    m1                  ;Endlos
.endif
;***********************************************************************************
;Hauptprogramm
;***********************************************************************************
main:   rcall   ad                  ;Spannung messen
        cpi     temp1,load          ;Max. Schwellwert erreicht? ca.14,2V 
        brcc    play                ;Wenn ja, zum Player springen
        cpi     temp1,lowbat        ;Sind wir unter 11,5 Volt ?
        brcs    blink               ;Wenn ja, LED1 blinken
        rcall   bar                 ;Mit den 8 Schwellwerten vergleichen und Bargraph bilden
        rcall   wait                ;Zeit verschwenden
        rjmp    main                ;Neue Runde
;***********************************************************************************
;Unterprogramme
;***********************************************************************************
        ;---------------------------------------------------------------------------
        ;Bargraph bilden
        ;---------------------------------------------------------------------------
bar:    ldi     zl,low(adwert*2)    ;Zeiger auf die Tab mit den Schwellwerten
        ldi     zh,high(adwert*2)
        clr     temp3               ;Hier werden fr leuchtende LEDs Einsen eingeschoben
bar1:   sec                         ;Eine LED leuchtet immer
        rol     temp3
        brmi     bar2               ;Wenn die Bar voll ist, Rcksprung
        lpm     temp2,z+            ;Ersten Vergleichswert holen
        cp      temp2,temp1         ;Wert aus Tab - AD-Wert
        brcs    bar1                ;Wenn AD-Wert <, wars das. Ansonsten eine 1 einschieben
bar2:   mov     leds,temp3          ;Ergebnis -> leds
        ret
        ;---------------------------------------------------------------------------
        ;Schwellwerte fr den Bargraph. 
        ;---------------------------------------------------------------------------
adwert: .db led2-1,led3-1,led4-1,led5-1,led6-1,led7-1,led8-1,0
        ;---------------------------------------------------------------------------
        ;ADC. 256 Messungen
        ;Der gemessene 10 Bit ADC-Wert wird so modifiziert, das in temp1 nur Spannungen
        ;von 10-15 Volt als 8-Bit Wert geliefert werden. Spannungen unter 10 Volt=0, ber
        ;15 Volt = 255
        ;---------------------------------------------------------------------------
ad:     clr     temp1               ;Lowbyte Messung
        clr     temp4               ;Highbyte
        clr     temp2               ;Zhler (temp2=256)
        clr     xh
ad1:    sbi     ADCSRA, ADSC        ;ADC starten 
w_adc:  sbic    ADCSRA, ADSC
        rjmp    w_adc               ;Auf die Messung warten
        in      xl,ADCL
        add     xh,xl               ;Durch addition 24Bit in xh(lsb), temp1 und temp4 
        in      xl,ADCH             ;xh wird wider verworfen (division /256)
        adc     temp1,xl
        adc     temp4,null
        dec     temp2               ;256 mal messen
        brne    ad1    
        ;---------------------------------------------------------------------------
        ;Den Bereich zwischen 10..15 Volt in temp1 auf 8 Bit begrenzen.
        ;Dazu Bit0 und 1 von ADCH auswerten.
        ;Der Bereich 0b10 0000 0000 - 0b10 1111 1111 (512-767) entspricht ~ 10-15V 
        ;---------------------------------------------------------------------------
        sbrc    temp4,0             ;ber 15 Volt?               
        ser     temp1               ;Wenn ja, 255 ins Ergebniss
        sbrs    temp4,1             ;Unter 10 Volt?
        clr     temp1               ;Wenn ja, 0 ins Ergebniss
        ret     
;****************************************************************************************
;Verschiedene Lauflichteffekte die beim erreichen von 14,2 Volt angezeigt werden
;****************************************************************************************
        ;---------------------------------------------------------------------------
        ;LED blinken bei Unterspannung. (Warnung vor Tiefentladung)
        ;---------------------------------------------------------------------------
blink:  clr     leds                ;LEDs aus
bl2:    rcall   wait
        inc     leds                ;Bit0 Toggeln
        andi    leds,0b1            ;Den Rest wegunden
        rcall   ad                  ;Zwischendurch messen
        cpi     temp1,lowbat        ;Bleiben wir noch ne runde?
        brcc    main                ;Wenn wir ber 11,5V sind, Rcksprung
        rjmp    bl2                 ;Ansonsten weiterblinken
        ;---------------------------------------------------------------------------
        ;Knight Rider
        ;---------------------------------------------------------------------------
        .if     player <3
        .if     player ==0
        .set    pt =0b00000001      ;Knight Rider mit einer LED
        .set    pt1=0b10000000      ;Startwerte fr hin und zurck
        .endif
        .if     player ==1
        .set    pt =0b00000011      ;Mit 2 LEDs
        .set    pt1=0b11000000
        .endif
        .if     player ==2
        .set    pt =0b00000111      ;Mit 3 LEDs
        .set    pt1=0b11100000
        .endif
play:   ldi     leds,pt             ;Startled(s)
kn1:    rcall   wait                ;Kurz warten
        rcall   ad                  ;Spannung messen
        cpi     temp1,load          ;Wenn unter Schwellwert, Rcksprung
        brcs    main
        lsl     leds                ;Die LED(s) wandern nach links
        brcc    kn1                 ;Solange nach links schieben bis die LED im Carry ankommt
        ldi     leds,pt1
kn2:    rcall   wait
        lsr     leds                ;Vorwrts Freunde, wir mssen zurck
        brcs    play
        rjmp    kn2
        .endif
        ;---------------------------------------------------------------------------
        ;Binrzhler
        ;---------------------------------------------------------------------------
        .if     player ==3
play:   clr     leds
bin1:   rcall   wait
        inc     leds
        rcall   ad
        cpi     temp1,load
        brcs    main
        rjmp    bin1
        .endif
        ;---------------------------------------------------------------------------
        ;Handy art. Die Bar fllt sich von links nach rechts
        ;---------------------------------------------------------------------------
        .if player ==4
play:   clr     leds            ;LEDs aus
ha:     rcall   wait
        rcall   ad              ;Zwischenduch messen
        cpi     temp1,load      ;Sind wir noch ber 14,2 Volt?
        brcs    main            ;Wenn nein, Rcksprung
        sec                     ;Eine LED in die Bar schieben
        rol     leds
        brcc    ha              ;War die Bar schon voll?
        rjmp    play            ;Wenn ja von vorne. 
        .endif
        ;---------------------------------------------------------------------------
        ;Tabelle
        ;---------------------------------------------------------------------------
        .if player ==5
play:   ldi     temp3,(tabende-tab)*2;Anzahl der Tabelleneintrge (Der Assembler kennt nur Words im Flash, deshalb *2)  
        ldi     zl,low(tab*2)   ;Zeiger auf Tab
        ldi     zh,high(tab*2)
tab1:   lpm     leds,z+         ;Tabelleneintrag nach LEDs holen
        rcall   ad              ;Zwischenduch messen
        cpi     temp1,load      ;Sind wir noch ber 14,2 Volt?
        brcs    main            ;Wenn nein, Rcksprung
        rcall   wait            ;Zeit vertrdeln
        dec     temp3           ;Anzahl Tabeintrge - 1
        brne    tab1            ;Ist noch was da? Wenn nein nchsten Tabeintrag bearbeiten
        rjmp    play            ;Ansonsten, das ganze von vorn
        ;---------------------------------------------------------------------------
        ;Tab mit dem Daumenkino
        ;---------------------------------------------------------------------------
tab:    .db 0b00000000,0b00011000
        .db 0b00111100,0b01111110
        .db 0b11111111,0b01111110
        .db 0b00111100,0b00011000
tabende:
        .endif
        ;---------------------------------------------------------------------------
        ;Zeit vertdeln
        ;---------------------------------------------------------------------------
wait:   ldi     temp1,delay
        mov     tick,temp1
wt:     tst     tick
        brne    wt
        ret
;***********************************************************************************
;Code Ende
;***********************************************************************************

